from __future__ import annotations

import logging
import pathlib
import threading
from typing import Callable, TypeVar

import grpc
from google.protobuf.any_pb2 import Any
from ni.measurementlink.discovery.v1.client import DiscoveryClient
from ni.panels.v1.panel_service_pb2 import (
    EnumeratePanelsRequest,
    GetValueRequest,
    SetValueRequest,
    StartPanelRequest,
    StopPanelRequest,
    TryGetValueRequest,
)
from ni.panels.v1.panel_service_pb2_grpc import PanelServiceStub
from ni.panels.v1.streamlit_panel_configuration_pb2 import StreamlitPanelConfiguration
from ni_grpc_extensions.channelpool import GrpcChannelPool
from typing_extensions import ParamSpec

from nipanel._convert import (
    from_any,
    to_any,
)

_P = ParamSpec("_P")
_T = TypeVar("_T")

_logger = logging.getLogger(__name__)

PANEL_SERVICE = "ni.panels.v1.PanelService"


class _PanelClient:
    def __init__(
        self,
        *,
        discovery_client: DiscoveryClient | None = None,
        grpc_channel_pool: GrpcChannelPool | None = None,
        grpc_channel: grpc.Channel | None = None,
    ) -> None:
        self._initialization_lock = threading.Lock()
        self._discovery_client = discovery_client
        self._grpc_channel_pool = grpc_channel_pool
        self._grpc_channel = grpc_channel
        self._stub: PanelServiceStub | None = None

    def start_streamlit_panel(
        self, panel_id: str, panel_script_path: pathlib.Path, python_interpreter_path: pathlib.Path
    ) -> str:

        panel_script_url = panel_script_path.absolute().as_uri()
        python_interpreter_url = python_interpreter_path.absolute().as_uri()
        streamlit_panel_configuration = StreamlitPanelConfiguration(
            panel_script_url=panel_script_url, python_interpreter_url=python_interpreter_url
        )
        panel_configuration_any = Any()
        panel_configuration_any.Pack(streamlit_panel_configuration)
        start_panel_request = StartPanelRequest(
            panel_id=panel_id, panel_configuration=panel_configuration_any
        )
        response = self._invoke_with_retry(self._get_stub().StartPanel, start_panel_request)
        return response.panel_url

    def stop_panel(self, panel_id: str, reset: bool) -> None:
        stop_panel_request = StopPanelRequest(panel_id=panel_id, reset=reset)
        self._invoke_with_retry(self._get_stub().StopPanel, stop_panel_request)

    def enumerate_panels(self) -> dict[str, tuple[str, list[str]]]:
        enumerate_panels_request = EnumeratePanelsRequest()
        response = self._invoke_with_retry(
            self._get_stub().EnumeratePanels, enumerate_panels_request
        )
        return {
            panel.panel_id: (panel.panel_url, list(panel.value_ids)) for panel in response.panels
        }

    def set_value(self, panel_id: str, value_id: str, value: object, notify: bool) -> None:
        new_any = to_any(value)
        set_value_request = SetValueRequest(
            panel_id=panel_id, value_id=value_id, value=new_any, notify=notify
        )
        self._invoke_with_retry(self._get_stub().SetValue, set_value_request)

    def get_value(self, panel_id: str, value_id: str) -> object:
        get_value_request = GetValueRequest(panel_id=panel_id, value_id=value_id)
        response = self._invoke_with_retry(self._get_stub().GetValue, get_value_request)
        return from_any(response.value)

    def try_get_value(self, panel_id: str, value_id: str) -> object | None:
        try_get_value_request = TryGetValueRequest(panel_id=panel_id, value_id=value_id)
        response = self._invoke_with_retry(self._get_stub().TryGetValue, try_get_value_request)
        if response.HasField("value"):
            return from_any(response.value)
        else:
            return None

    def _get_stub(self) -> PanelServiceStub:
        if self._stub is None:
            if self._grpc_channel is not None:
                self._stub = PanelServiceStub(self._grpc_channel)
            else:
                with self._initialization_lock:
                    if self._grpc_channel_pool is None:
                        _logger.debug("Creating unshared GrpcChannelPool.")
                        self._grpc_channel_pool = GrpcChannelPool()
                    if self._discovery_client is None:
                        _logger.debug("Creating unshared DiscoveryClient.")
                        self._discovery_client = DiscoveryClient(
                            grpc_channel_pool=self._grpc_channel_pool
                        )

                    service_location = self._discovery_client.resolve_service(
                        provided_interface=PANEL_SERVICE
                    )
                    channel = self._grpc_channel_pool.get_channel(service_location.insecure_address)
                    self._stub = PanelServiceStub(channel)
        return self._stub

    def _invoke_with_retry(
        self, method: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs
    ) -> _T:
        """Invoke a gRPC method with retry logic."""
        try:
            return method(*args, **kwargs)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE or e.code() == grpc.StatusCode.UNKNOWN:
                # if the service is unavailable, we can retry the connection
                self._stub = None
                return method(*args, **kwargs)
            raise



"""Controls for nipanel."""

from nipanel.controls._enum_selectbox import enum_selectbox
from nipanel.controls._flag_checkboxes import flag_checkboxes

__all__ = ["enum_selectbox", "flag_checkboxes"]



from pathlib import Path
from typing import cast

import streamlit as st

from nipanel._convert import is_supported_type
from nipanel._panel_value_accessor import PanelValueAccessor
from nipanel._streamlit_panel import StreamlitPanel
from nipanel.streamlit_refresh import initialize_refresh_component

PANEL_ACCESSOR_KEY = "StreamlitPanelValueAccessor"


def create_streamlit_panel(streamlit_script_path: Path, panel_id: str = "") -> StreamlitPanel:
    """Create a Streamlit panel with the specified script path.

    This function initializes a Streamlit panel using the provided script path. By default, it
    derives the panel ID from the script's path, which it expects to be a valid Streamlit script.
    For example, if the value for streamlit_script_path is "c:/example/some_example.py", then the
    panel's ID becomes "some_example". Alternatively, you can specify a custom panel_id.

    Use this function when you want to create a new panel instance to use in a Streamlit
    application. Do not call this function from within a Streamlit script.

    Args:
        streamlit_script_path: The file path of the Streamlit script to be used for the panel.
        panel_id: Optional custom panel ID. If not provided, it will be derived from the script
            path.

    Returns:
        A StreamlitPanel instance initialized with the given panel ID.
    """
    if st.get_option("server.baseUrlPath") != "":
        raise RuntimeError(
            "nipanel.create_panel() should not be called from a Streamlit script. Call nipanel.get_panel_accessor() instead."
        )
    if not isinstance(streamlit_script_path, Path):
        raise TypeError("The provided script path must be a pathlib.Path instance.")

    if streamlit_script_path.suffix != ".py":
        raise ValueError(
            "The provided script path must be a valid Streamlit script ending with '.py'."
        )

    if not panel_id:
        panel_id = streamlit_script_path.stem

    return StreamlitPanel(panel_id, streamlit_script_path)


def get_streamlit_panel_accessor() -> PanelValueAccessor:
    """Initialize and return the Streamlit panel value accessor.

    This function retrieves the Streamlit panel value accessor for the current Streamlit script.
    This function should only be called from within a Streamlit script. The accessor will be cached
    in the Streamlit session state to ensure that it is reused across reruns of the script.

    Returns:
        A PanelValueAccessor instance for the current panel.
    """
    if st.get_option("server.baseUrlPath") == "":
        raise RuntimeError(
            "nipanel.get_panel_accessor() should only be called from a Streamlit script. Call nipanel.create_panel() instead."
        )

    if PANEL_ACCESSOR_KEY not in st.session_state:
        st.session_state[PANEL_ACCESSOR_KEY] = _initialize_panel_from_base_path()

    panel = cast(PanelValueAccessor, st.session_state[PANEL_ACCESSOR_KEY])
    _sync_session_state(panel)
    refresh_component = initialize_refresh_component(panel.panel_id)
    refresh_component()
    return panel


def _initialize_panel_from_base_path() -> PanelValueAccessor:
    """Validate and parse the Streamlit base URL path and return a PanelValueAccessor."""
    base_url_path = st.get_option("server.baseUrlPath")
    if not base_url_path.startswith("/"):
        raise ValueError("Invalid or missing Streamlit server.baseUrlPath option.")
    panel_id = base_url_path.split("/")[-1]
    if not panel_id:
        raise ValueError(f"Panel ID is empty in baseUrlPath: '{base_url_path}'")
    return PanelValueAccessor(
        panel_id=panel_id,
        notify_on_set_value=False,
    )


def _sync_session_state(panel: PanelValueAccessor) -> None:
    """Automatically read keyed control values from the session state."""
    for key in st.session_state.keys():
        value = st.session_state[key]
        if is_supported_type(value):
            panel.set_value_if_changed(str(key), value)



"""Initializes a refresh component for Streamlit."""

from __future__ import annotations

import threading

from ni.measurementlink.discovery.v1.client import DiscoveryClient
from ni_grpc_extensions.channelpool import GrpcChannelPool
from streamlit.components.v1 import declare_component
from streamlit.components.v1.custom_component import CustomComponent


_grpc_client_lock = threading.RLock()
_panel_service_proxy_location: str | None = None


def initialize_refresh_component(panel_id: str) -> CustomComponent:
    """Initialize a refresh component to the Streamlit app."""
    proxy_base_address = _get_or_resolve_proxy()
    component_url = f"http://{proxy_base_address}/panel-service/refresh/{panel_id}"
    _refresh_component_func = declare_component(
        "panelRefreshComponent",
        url=component_url,
    )

    return _refresh_component_func


def _get_or_resolve_proxy() -> str:
    with _grpc_client_lock:
        global _panel_service_proxy_location
        if _panel_service_proxy_location is None:
            with GrpcChannelPool() as grpc_channel_pool:
                discovery_client = DiscoveryClient(grpc_channel_pool=grpc_channel_pool)
                service_location = discovery_client.resolve_service(
                    provided_interface="ni.http1.proxy",
                    service_class="",
                )
                _panel_service_proxy_location = service_location.insecure_address

        return _panel_service_proxy_location



from __future__ import annotations

import collections
import enum
from abc import ABC
from typing import TypeVar, overload

import grpc
import hightime as ht
import nitypes.bintime as bt
from ni.measurementlink.discovery.v1.client import DiscoveryClient
from ni_grpc_extensions.channelpool import GrpcChannelPool
from nitypes.time import convert_datetime

from nipanel._panel_client import _PanelClient

_T = TypeVar("_T")


class PanelValueAccessor(ABC):
    """This class allows you to access values for a panel's controls."""

    __slots__ = [
        "_panel_client",
        "_panel_id",
        "_notify_on_set_value",
        "_last_values",
        "__weakref__",
    ]

    def __init__(
        self,
        *,
        panel_id: str,
        notify_on_set_value: bool = True,
        discovery_client: DiscoveryClient | None = None,
        grpc_channel_pool: GrpcChannelPool | None = None,
        grpc_channel: grpc.Channel | None = None,
    ) -> None:
        """Initialize the accessor."""
        self._panel_client = _PanelClient(
            discovery_client=discovery_client,
            grpc_channel_pool=grpc_channel_pool,
            grpc_channel=grpc_channel,
        )
        self._panel_id = panel_id
        self._notify_on_set_value = notify_on_set_value
        self._last_values: collections.defaultdict[str, object] = collections.defaultdict(
            lambda: object()
        )

    @property
    def panel_id(self) -> str:
        """Read-only accessor for the panel ID."""
        return self._panel_id

    @overload
    def get_value(self, value_id: str) -> object: ...

    @overload
    def get_value(self, value_id: str, default_value: _T) -> _T: ...

    def get_value(self, value_id: str, default_value: _T | None = None) -> _T | object:
        """Get the value for a control on the panel with an optional default value.

        Args:
            value_id: The id of the value
            default_value: The default value to return if the value is not set

        Returns:
            The value, or the default value if not set. The returned value will
            have the same type as default_value, if one was provided.

        Raises:
            KeyError: If the value is not set and no default value is provided
        """
        value = self._panel_client.try_get_value(self._panel_id, value_id)
        if value is None:
            if default_value is not None:
                return default_value
            raise KeyError(f"Value with id '{value_id}' not found on panel '{self._panel_id}'.")

        if default_value is not None and not isinstance(value, type(default_value)):
            if isinstance(default_value, enum.Enum):
                enum_type = type(default_value)
                return enum_type(value)

            # The grpc converter always converts PrecisionTimestamp into bt.DateTime, so
            # we need to handle the case where they provide an ht.datetime default by
            # converting to hightime.
            if isinstance(default_value, ht.datetime) and isinstance(value, bt.DateTime):
                return convert_datetime(ht.datetime, value)

            # lists are allowed to not match, since sets and tuples are converted to lists
            if not isinstance(value, list):
                raise TypeError(
                    f"Value type {type(value).__name__} does not match default value type {type(default_value).__name__}."
                )

        return value

    def set_value(self, value_id: str, value: object) -> None:
        """Set the value for a control on the panel.

        Args:
            value_id: The id of the value
            value: The value
        """
        if isinstance(value, enum.Enum):
            value = value.value

        self._panel_client.set_value(
            self._panel_id, value_id, value, notify=self._notify_on_set_value
        )
        self._last_values[value_id] = value

    def set_value_if_changed(self, value_id: str, value: object) -> None:
        """Set the value for a control on the panel only if it has changed since the last call.

        This method helps reduce unnecessary updates when the value hasn't changed.

        Args:
            value_id: The id of the value
            value: The value to set
        """
        if value != self._last_values[value_id]:
            self.set_value(value_id, value)



from __future__ import annotations

import sys
from pathlib import Path
from typing import final

import grpc
from ni.measurementlink.discovery.v1.client import DiscoveryClient
from ni_grpc_extensions.channelpool import GrpcChannelPool

from nipanel._panel_value_accessor import PanelValueAccessor


@final
class StreamlitPanel(PanelValueAccessor):
    """This class allows you to open a Streamlit panel and specify values for its controls."""

    __slots__ = ["_panel_script_path", "_panel_url"]

    def __init__(
        self,
        panel_id: str,
        panel_script_path: Path,
        *,
        discovery_client: DiscoveryClient | None = None,
        grpc_channel_pool: GrpcChannelPool | None = None,
        grpc_channel: grpc.Channel | None = None,
    ) -> None:
        """Create a panel using a Streamlit script for the user interface.

        Args:
            panel_id: A unique identifier for the panel.
            panel_script_path: The file path of the Streamlit script.
            discovery_client: An optional DiscoveryClient for service discovery.
            grpc_channel_pool: An optional GrpcChannelPool for managing gRPC channels.
            grpc_channel: An optional gRPC channel to use for communication with the panel service.

        Returns:
            A new StreamlitPanel instance.
        """
        super().__init__(
            panel_id=panel_id,
            discovery_client=discovery_client,
            grpc_channel_pool=grpc_channel_pool,
            grpc_channel=grpc_channel,
        )
        self._panel_script_path = panel_script_path
        python_path = self._get_python_path()
        self._panel_url = self._panel_client.start_streamlit_panel(
            panel_id, panel_script_path, python_path
        )

    @property
    def panel_script_path(self) -> Path:
        """Read-only accessor for the streamlit script file path."""
        return self._panel_script_path

    @property
    def panel_url(self) -> str:
        """Read-only accessor for the panel's streamlit webpage URL."""
        return self._panel_url

    def _get_python_path(self) -> Path:
        """Get the Python interpreter path for the panel that ensures the same environment."""
        if sys.executable is None or sys.executable == "":
            raise RuntimeError("Python environment not found")
        if getattr(sys, "frozen", False):
            raise RuntimeError("Panel cannot be used in a frozen application (e.g., PyInstaller).")

        if sys.prefix != sys.base_prefix:
            # If we're in a virtual environment, build the path to the Python executable
            # On Linux: .venv/bin/python, On Windows: .venv\Scripts\python.exe
            if sys.platform.startswith("win"):
                python_executable = "python.exe"
                bin_dir = "Scripts"
            else:
                python_executable = "python"
                bin_dir = "bin"

            # Construct path to the Python in the virtual environment based on sys.prefix
            python_path = Path(sys.prefix) / bin_dir / python_executable

            # Fall back to sys.executable if the constructed path doesn't exist
            if not python_path.exists():
                python_path = Path(sys.executable).resolve()
        else:
            # If not in a .venv environment, use sys.executable
            python_path = Path(sys.executable).resolve()

        if python_path.is_relative_to(Path(sys.prefix)) is False:
            # Ensure the Python path is within the current environment
            raise RuntimeError(
                f"Python path '{python_path}' does not match the current environment prefix '{sys.prefix}'."
            )

        return python_path



"""Functions and classes to convert types between Python and protobuf."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Collection
from typing import Generic, Type, TypeVar

from google.protobuf import any_pb2
from google.protobuf.message import Message

_TItemType = TypeVar("_TItemType")
_TPythonType = TypeVar("_TPythonType")
_TProtobufType = TypeVar("_TProtobufType", bound=Message)


class Converter(Generic[_TPythonType, _TProtobufType], ABC):
    """A class that defines how to convert between Python objects and protobuf Any messages."""

    @property
    @abstractmethod
    def python_type(self) -> type:
        """The Python type that this converter handles."""

    @property
    def python_typename(self) -> str:
        """The Python type name that this converter handles."""
        return f"{self.python_type.__module__}.{self.python_type.__name__}"

    @property
    @abstractmethod
    def protobuf_message(self) -> Type[_TProtobufType]:
        """The type-specific protobuf message for the Python type."""

    @property
    def protobuf_typename(self) -> str:
        """The protobuf name for the type."""
        return self.protobuf_message.DESCRIPTOR.full_name  # type: ignore[no-any-return]

    def to_protobuf_any(self, python_value: _TPythonType) -> any_pb2.Any:
        """Convert the Python object to its type-specific message and pack it as any_pb2.Any."""
        message = self.to_protobuf_message(python_value)
        as_any = any_pb2.Any()
        as_any.Pack(message)
        return as_any

    @abstractmethod
    def to_protobuf_message(self, python_value: _TPythonType) -> _TProtobufType:
        """Convert the Python object to its type-specific message."""

    def to_python(self, protobuf_value: any_pb2.Any) -> _TPythonType:
        """Convert the protobuf Any message to its matching Python type."""
        protobuf_message = self.protobuf_message()
        did_unpack = protobuf_value.Unpack(protobuf_message)
        if not did_unpack:
            raise ValueError(f"Failed to unpack Any with type '{protobuf_value.TypeName()}'")
        return self.to_python_value(protobuf_message)

    @abstractmethod
    def to_python_value(self, protobuf_message: _TProtobufType) -> _TPythonType:
        """Convert the protobuf wrapper message to its matching Python type."""


class CollectionConverter(
    Generic[_TItemType, _TProtobufType],
    Converter[Collection[_TItemType], _TProtobufType],
    ABC,
):
    """A converter between a collection of Python objects and protobuf Any messages."""

    @property
    @abstractmethod
    def item_type(self) -> type:
        """The Python item type that this converter handles."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return Collection

    @property
    def python_typename(self) -> str:
        """The Python type name that this converter handles."""
        return "{}[{}]".format(
            f"{Collection.__module__}.{Collection.__name__}",
            f"{self.item_type.__module__}.{self.item_type.__name__}",
        )


class CollectionConverter2D(
    Generic[_TItemType, _TProtobufType],
    Converter[Collection[Collection[_TItemType]], _TProtobufType],
    ABC,
):
    """A converter between a 2D collection of Python objects and protobuf Any messages."""

    @property
    @abstractmethod
    def item_type(self) -> type:
        """The Python item type that this converter handles."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return Collection

    @property
    def python_typename(self) -> str:
        """The Python type name that this converter handles."""
        return "{}[{}[{}]]".format(
            f"{Collection.__module__}.{Collection.__name__}",
            f"{Collection.__module__}.{Collection.__name__}",
            f"{self.item_type.__module__}.{self.item_type.__name__}",
        )



"""Classes to convert between builtin Python scalars and containers."""

import datetime as dt
from typing import Type

from google.protobuf import duration_pb2, timestamp_pb2, wrappers_pb2

from nipanel.converters import Converter


class BoolConverter(Converter[bool, wrappers_pb2.BoolValue]):
    """A converter for boolean types."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return bool

    @property
    def protobuf_message(self) -> Type[wrappers_pb2.BoolValue]:
        """The type-specific protobuf message for the Python type."""
        return wrappers_pb2.BoolValue

    def to_protobuf_message(self, python_value: bool) -> wrappers_pb2.BoolValue:
        """Convert the Python bool to a protobuf wrappers_pb2.BoolValue."""
        return self.protobuf_message(value=python_value)

    def to_python_value(self, protobuf_message: wrappers_pb2.BoolValue) -> bool:
        """Convert the protobuf message to a Python bool."""
        return protobuf_message.value


class BytesConverter(Converter[bytes, wrappers_pb2.BytesValue]):
    """A converter for byte string types."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return bytes

    @property
    def protobuf_message(self) -> Type[wrappers_pb2.BytesValue]:
        """The type-specific protobuf message for the Python type."""
        return wrappers_pb2.BytesValue

    def to_protobuf_message(self, python_value: bytes) -> wrappers_pb2.BytesValue:
        """Convert the Python bytes string to a protobuf wrappers_pb2.BytesValue."""
        return self.protobuf_message(value=python_value)

    def to_python_value(self, protobuf_message: wrappers_pb2.BytesValue) -> bytes:
        """Convert the protobuf message to a Python bytes string."""
        return protobuf_message.value


class FloatConverter(Converter[float, wrappers_pb2.DoubleValue]):
    """A converter for floating point types."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return float

    @property
    def protobuf_message(self) -> Type[wrappers_pb2.DoubleValue]:
        """The type-specific protobuf message for the Python type."""
        return wrappers_pb2.DoubleValue

    def to_protobuf_message(self, python_value: float) -> wrappers_pb2.DoubleValue:
        """Convert the Python float to a protobuf wrappers_pb2.DoubleValue."""
        return self.protobuf_message(value=python_value)

    def to_python_value(self, protobuf_message: wrappers_pb2.DoubleValue) -> float:
        """Convert the protobuf message to a Python float."""
        return protobuf_message.value


class IntConverter(Converter[int, wrappers_pb2.Int64Value]):
    """A converter for integer types."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return int

    @property
    def protobuf_message(self) -> Type[wrappers_pb2.Int64Value]:
        """The type-specific protobuf message for the Python type."""
        return wrappers_pb2.Int64Value

    def to_protobuf_message(self, python_value: int) -> wrappers_pb2.Int64Value:
        """Convert the Python int to a protobuf wrappers_pb2.Int64Value."""
        return self.protobuf_message(value=python_value)

    def to_python_value(self, protobuf_message: wrappers_pb2.Int64Value) -> int:
        """Convert the protobuf message to a Python int."""
        return protobuf_message.value


class StrConverter(Converter[str, wrappers_pb2.StringValue]):
    """A converter for text string types."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return str

    @property
    def protobuf_message(self) -> Type[wrappers_pb2.StringValue]:
        """The type-specific protobuf message for the Python type."""
        return wrappers_pb2.StringValue

    def to_protobuf_message(self, python_value: str) -> wrappers_pb2.StringValue:
        """Convert the Python str to a protobuf wrappers_pb2.StringValue."""
        return self.protobuf_message(value=python_value)

    def to_python_value(self, protobuf_message: wrappers_pb2.StringValue) -> str:
        """Convert the protobuf message to a Python string."""
        return protobuf_message.value


class DTDateTimeConverter(Converter[dt.datetime, timestamp_pb2.Timestamp]):
    """A converter for datetime.datetime types."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return dt.datetime

    @property
    def protobuf_message(self) -> Type[timestamp_pb2.Timestamp]:
        """The type-specific protobuf message for the Python type."""
        return timestamp_pb2.Timestamp

    def to_protobuf_message(self, python_value: dt.datetime) -> timestamp_pb2.Timestamp:
        """Convert the Python dt.datetime to a protobuf timestamp_pb2.Timestamp."""
        ts = self.protobuf_message()
        ts.FromDatetime(python_value)
        return ts

    def to_python_value(self, protobuf_message: timestamp_pb2.Timestamp) -> dt.datetime:
        """Convert the protobuf timestamp_pb2.Timestamp to a Python dt.datetime."""
        return protobuf_message.ToDatetime()


class DTTimeDeltaConverter(Converter[dt.timedelta, duration_pb2.Duration]):
    """A converter for datetime.timedelta types."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return dt.timedelta

    @property
    def protobuf_message(self) -> Type[duration_pb2.Duration]:
        """The type-specific protobuf message for the Python type."""
        return duration_pb2.Duration

    def to_protobuf_message(self, python_value: dt.timedelta) -> duration_pb2.Duration:
        """Convert the Python dt.timedelta to a protobuf duration_pb2.Duration."""
        dur = self.protobuf_message()
        dur.FromTimedelta(python_value)
        return dur

    def to_python_value(self, protobuf_message: duration_pb2.Duration) -> dt.timedelta:
        """Convert the protobuf timestamp_pb2.Timestamp to a Python dt.timedelta."""
        return protobuf_message.ToTimedelta()



"""Functions to convert between different data formats."""

from __future__ import annotations

import logging
from collections.abc import Collection
from typing import Any, Iterable

from google.protobuf import any_pb2
from nitypes.waveform import AnalogWaveform, ComplexWaveform

from nipanel.converters import Converter
from nipanel.converters.builtin import (
    BoolConverter,
    BytesConverter,
    DTDateTimeConverter,
    DTTimeDeltaConverter,
    FloatConverter,
    IntConverter,
    StrConverter,
)
from nipanel.converters.protobuf_types import (
    BTDateTimeConverter,
    BoolCollectionConverter,
    BytesCollectionConverter,
    DigitalWaveformConverter,
    Double2DArrayConverter,
    DoubleAnalogWaveformConverter,
    DoubleComplexWaveformConverter,
    DoubleSpectrumConverter,
    FloatCollectionConverter,
    HTDateTimeConverter,
    Int16AnalogWaveformConverter,
    Int16ComplexWaveformConverter,
    IntCollectionConverter,
    ScalarConverter,
    StrCollectionConverter,
    VectorConverter,
)

_logger = logging.getLogger(__name__)

# FFV -- consider adding a RegisterConverter mechanism
_CONVERTIBLE_TYPES: list[Converter[Any, Any]] = [
    # Built-in Types
    BoolConverter(),
    BytesConverter(),
    FloatConverter(),
    IntConverter(),
    StrConverter(),
    DTDateTimeConverter(),
    DTTimeDeltaConverter(),
    # Protobuf Types
    BTDateTimeConverter(),
    BoolCollectionConverter(),
    BytesCollectionConverter(),
    DigitalWaveformConverter(),
    Double2DArrayConverter(),
    DoubleAnalogWaveformConverter(),
    DoubleComplexWaveformConverter(),
    DoubleSpectrumConverter(),
    FloatCollectionConverter(),
    HTDateTimeConverter(),
    Int16AnalogWaveformConverter(),
    Int16ComplexWaveformConverter(),
    IntCollectionConverter(),
    StrCollectionConverter(),
    ScalarConverter(),
    VectorConverter(),
]

_CONVERTIBLE_COLLECTION_TYPES = {
    frozenset,
    list,
    set,
    tuple,
}

_CONVERTER_FOR_PYTHON_TYPE = {entry.python_typename: entry for entry in _CONVERTIBLE_TYPES}
_CONVERTER_FOR_GRPC_TYPE = {entry.protobuf_typename: entry for entry in _CONVERTIBLE_TYPES}
_SUPPORTED_PYTHON_TYPES = _CONVERTER_FOR_PYTHON_TYPE.keys()


def to_any(python_value: object) -> any_pb2.Any:
    """Convert a Python object to a protobuf Any."""
    best_matching_type = _get_best_matching_type(python_value)
    converter = _CONVERTER_FOR_PYTHON_TYPE[best_matching_type]
    return converter.to_protobuf_any(python_value)


def _get_best_matching_type(python_value: object) -> str:
    underlying_parents = type(python_value).mro()  # This covers enum.IntEnum and similar
    additional_info_string = _get_additional_type_info_string(python_value)

    container_types = []
    value_is_collection = any(_CONVERTIBLE_COLLECTION_TYPES.intersection(underlying_parents))
    # Variable to use when traversing down through collection types.
    working_python_value = python_value
    while value_is_collection:
        # Assume Sized -- Generators not supported, callers must use list(), set(), ... as desired
        if not isinstance(working_python_value, Collection):
            raise TypeError()
        if len(working_python_value) == 0:
            underlying_parents = type(None).mro()
            value_is_collection = False
        else:
            # Assume homogenous -- collections of mixed-types not supported
            visitor = iter(working_python_value)

            # Store off the first element. If it's a container, we'll need it in the next while
            # loop iteration.
            working_python_value = next(visitor)
            underlying_parents = type(working_python_value).mro()

            # If this element is a collection, we want to continue traversing. Once we find a
            # non-collection, underlying_parents will refer to the candidates for the non-
            # collection type.
            value_is_collection = any(
                _CONVERTIBLE_COLLECTION_TYPES.intersection(underlying_parents)
            )
        container_types.append(Collection)

    best_matching_type = None
    candidates = _get_candidate_strings(underlying_parents)
    for candidate in candidates:
        python_typename = _create_python_typename(
            candidate, container_types, additional_info_string
        )
        if python_typename not in _SUPPORTED_PYTHON_TYPES:
            continue
        best_matching_type = python_typename
        break

    if not best_matching_type:
        payload_type = underlying_parents[0]
        raise TypeError(
            f"Unsupported type: ({container_types}, {payload_type}) with parents "
            f"{underlying_parents}.\n\nSupported types are: {_SUPPORTED_PYTHON_TYPES}"
            f"\n\nAdditional type info: {additional_info_string}"
        )
    _logger.debug(f"Best matching type for '{repr(python_value)}' resolved to {best_matching_type}")
    return best_matching_type


def from_any(protobuf_any: any_pb2.Any) -> object:
    """Convert a protobuf Any to a Python object."""
    if not isinstance(protobuf_any, any_pb2.Any):
        raise ValueError(f"Unexpected type: {type(protobuf_any)}")

    underlying_typename = protobuf_any.TypeName()
    _logger.debug(f"Unpacking type '{underlying_typename}'")

    converter = _CONVERTER_FOR_GRPC_TYPE[underlying_typename]
    return converter.to_python(protobuf_any)


def is_supported_type(value: object) -> bool:
    """Check if a given Python value can be converted to protobuf Any."""
    try:
        _get_best_matching_type(value)
        return True
    except TypeError:
        return False


def _get_candidate_strings(candidates: Iterable[type]) -> list[str]:
    candidate_names = []
    for candidate in candidates:
        candidate_names.append(f"{candidate.__module__}.{candidate.__name__}")
    return candidate_names


def _create_python_typename(
    candidate_name: str, container_types: Iterable[type], additional_info: str
) -> str:
    name = candidate_name
    if additional_info:
        name = f"{name}[{additional_info}]"
    for container_type in container_types:
        name = f"{container_type.__module__}.{container_type.__name__}[{name}]"
    return name


def _get_additional_type_info_string(python_value: object) -> str:
    if isinstance(python_value, AnalogWaveform):
        return str(python_value.dtype)
    elif isinstance(python_value, ComplexWaveform):
        return str(python_value.dtype)
    else:
        return ""



"""A selectbox that allows selecting an Enum value."""

from enum import Enum
from typing import Any, Callable, TypeVar

import streamlit as st

from nipanel._panel_value_accessor import PanelValueAccessor

TEnumType = TypeVar("TEnumType", bound=Enum)


def enum_selectbox(
    panel: PanelValueAccessor,
    label: str,
    value: TEnumType,
    key: str,
    disabled: bool = False,
    format_func: Callable[[Any], str] = lambda x: x[0],
) -> TEnumType:
    """Create a selectbox for an Enum.

    The selectbox will display the names of all the enum values, and when a value is selected,
    that value will be stored in the panel under the specified key.

    Args:
        panel: The panel
        label: Label to display for the selectbox
        value: The default enum value to select (also determines the specific enum type)
        key: Key to use for storing the enum value in the panel

    Returns:
        The selected enum value of the same specific enum subclass as the input value
    """
    enum_class = type(value)
    if not issubclass(enum_class, Enum):
        raise TypeError(f"Expected an Enum type, got {type(value)}")

    options = [(e.name, e.value) for e in enum_class]

    default_index = 0
    if value is not None:
        for i, (name, _) in enumerate(options):
            if name == value.name:
                default_index = i
                break

    box_tuple = st.selectbox(
        label, options=options, format_func=format_func, index=default_index, disabled=disabled
    )
    enum_value = enum_class[box_tuple[0]]
    panel.set_value_if_changed(key, enum_value)
    return enum_value



"""Classes to convert between measurement specific protobuf types and containers."""

from __future__ import annotations

from collections.abc import Collection
from typing import Any, Type, Union

import hightime as ht
import nitypes.bintime as bt
import numpy as np
from ni.protobuf.types import (
    array_pb2,
    precision_timestamp_pb2,
    precision_timestamp_conversion,
    scalar_conversion,
    scalar_pb2,
    vector_pb2,
    vector_conversion,
    waveform_conversion,
    waveform_pb2,
)
from nitypes.complex import ComplexInt32Base
from nitypes.scalar import Scalar
from nitypes.vector import Vector
from nitypes.waveform import AnalogWaveform, ComplexWaveform, DigitalWaveform, Spectrum
from typing_extensions import TypeAlias

from nipanel.converters import Converter, CollectionConverter, CollectionConverter2D

_AnyScalarType: TypeAlias = Union[bool, int, float, str]


class BoolCollectionConverter(CollectionConverter[bool, array_pb2.BoolArray]):
    """A converter for a Collection of bools."""

    @property
    def item_type(self) -> type:
        """The Python type that this converter handles."""
        return bool

    @property
    def protobuf_message(self) -> Type[array_pb2.BoolArray]:
        """The type-specific protobuf message for the Python type."""
        return array_pb2.BoolArray

    def to_protobuf_message(self, python_value: Collection[bool]) -> array_pb2.BoolArray:
        """Convert the collection of bools to array_pb2.BoolArray."""
        return self.protobuf_message(values=python_value)

    def to_python_value(self, protobuf_message: array_pb2.BoolArray) -> Collection[bool]:
        """Convert the protobuf message to a Python collection of bools."""
        return list(protobuf_message.values)


class BytesCollectionConverter(CollectionConverter[bytes, array_pb2.BytesArray]):
    """A converter for a Collection of byte strings."""

    @property
    def item_type(self) -> type:
        """The Python type that this converter handles."""
        return bytes

    @property
    def protobuf_message(self) -> Type[array_pb2.BytesArray]:
        """The type-specific protobuf message for the Python type."""
        return array_pb2.BytesArray

    def to_protobuf_message(self, python_value: Collection[bytes]) -> array_pb2.BytesArray:
        """Convert the collection of byte strings to array_pb2.BytesArray."""
        return self.protobuf_message(values=python_value)

    def to_python_value(self, protobuf_message: array_pb2.BytesArray) -> Collection[bytes]:
        """Convert the protobuf message to a Python collection of byte strings."""
        return list(protobuf_message.values)


class FloatCollectionConverter(CollectionConverter[float, array_pb2.DoubleArray]):
    """A converter for a Collection of floats."""

    @property
    def item_type(self) -> type:
        """The Python type that this converter handles."""
        return float

    @property
    def protobuf_message(self) -> Type[array_pb2.DoubleArray]:
        """The type-specific protobuf message for the Python type."""
        return array_pb2.DoubleArray

    def to_protobuf_message(self, python_value: Collection[float]) -> array_pb2.DoubleArray:
        """Convert the collection of floats to array_pb2.DoubleArray."""
        return self.protobuf_message(values=python_value)

    def to_python_value(self, protobuf_message: array_pb2.DoubleArray) -> Collection[float]:
        """Convert the protobuf message to a Python collection of floats."""
        return list(protobuf_message.values)


class IntCollectionConverter(CollectionConverter[int, array_pb2.SInt64Array]):
    """A converter for a Collection of integers."""

    @property
    def item_type(self) -> type:
        """The Python type that this converter handles."""
        return int

    @property
    def protobuf_message(self) -> Type[array_pb2.SInt64Array]:
        """The type-specific protobuf message for the Python type."""
        return array_pb2.SInt64Array

    def to_protobuf_message(self, python_value: Collection[int]) -> array_pb2.SInt64Array:
        """Convert the collection of integers to array_pb2.SInt64Array."""
        return self.protobuf_message(values=python_value)

    def to_python_value(self, protobuf_message: array_pb2.SInt64Array) -> Collection[int]:
        """Convert the protobuf message to a Python collection of integers."""
        return list(protobuf_message.values)


class StrCollectionConverter(CollectionConverter[str, array_pb2.StringArray]):
    """A converter for a Collection of strings."""

    @property
    def item_type(self) -> type:
        """The Python type that this converter handles."""
        return str

    @property
    def protobuf_message(self) -> Type[array_pb2.StringArray]:
        """The type-specific protobuf message for the Python type."""
        return array_pb2.StringArray

    def to_protobuf_message(self, python_value: Collection[str]) -> array_pb2.StringArray:
        """Convert the collection of strings to array_pb2.StringCollection."""
        return self.protobuf_message(values=python_value)

    def to_python_value(self, protobuf_message: array_pb2.StringArray) -> Collection[str]:
        """Convert the protobuf message to a Python collection of strings."""
        return list(protobuf_message.values)


class Double2DArrayConverter(CollectionConverter2D[float, array_pb2.Double2DArray]):
    """A converter between Collection[Collection[float]] and Double2DArray."""

    @property
    def item_type(self) -> type:
        """The Python item type that this converter handles."""
        return float

    @property
    def protobuf_message(self) -> Type[array_pb2.Double2DArray]:
        """The type-specific protobuf message for the Python type."""
        return array_pb2.Double2DArray

    def to_protobuf_message(
        self, python_value: Collection[Collection[float]]
    ) -> array_pb2.Double2DArray:
        """Convert the Python Collection[Collection[float]] to a protobuf Double2DArray."""
        rows = len(python_value)
        if rows:
            visitor = iter(python_value)
            first_subcollection = next(visitor)
            columns = len(first_subcollection)
        else:
            columns = 0
        if not all(len(subcollection) == columns for subcollection in python_value):
            raise ValueError("All subcollections must have the same length.")

        # Create a flat list in row major order.
        flat_list = [item for subcollection in python_value for item in subcollection]
        return array_pb2.Double2DArray(rows=rows, columns=columns, data=flat_list)

    def to_python_value(
        self, protobuf_message: array_pb2.Double2DArray
    ) -> Collection[Collection[float]]:
        """Convert the protobuf Double2DArray to a Python Collection[Collection[float]]."""
        if not protobuf_message.data:
            return []
        if len(protobuf_message.data) % protobuf_message.columns != 0:
            raise ValueError("The length of the data list must be divisible by num columns.")

        # Convert from a flat list in row major order into a list of lists.
        list_of_lists = []
        for i in range(0, len(protobuf_message.data), protobuf_message.columns):
            row = protobuf_message.data[i : i + protobuf_message.columns]
            list_of_lists.append(row)

        return list_of_lists


class DoubleAnalogWaveformConverter(
    Converter[AnalogWaveform[np.float64], waveform_pb2.DoubleAnalogWaveform]
):
    """A converter for AnalogWaveform types with double-precision data."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return AnalogWaveform

    @property
    def python_typename(self) -> str:
        """The Python type name that this converter handles."""
        base_typename = super().python_typename
        return f"{base_typename}[float64]"

    @property
    def protobuf_message(self) -> Type[waveform_pb2.DoubleAnalogWaveform]:
        """The type-specific protobuf message for the Python type."""
        return waveform_pb2.DoubleAnalogWaveform

    def to_protobuf_message(
        self, python_value: AnalogWaveform[np.float64]
    ) -> waveform_pb2.DoubleAnalogWaveform:
        """Convert the Python AnalogWaveform to a protobuf DoubleAnalogWaveform."""
        return waveform_conversion.float64_analog_waveform_to_protobuf(python_value)

    def to_python_value(
        self, protobuf_message: waveform_pb2.DoubleAnalogWaveform
    ) -> AnalogWaveform[np.float64]:
        """Convert the protobuf DoubleAnalogWaveform to a Python AnalogWaveform."""
        return waveform_conversion.float64_analog_waveform_from_protobuf(protobuf_message)


class Int16AnalogWaveformConverter(
    Converter[AnalogWaveform[np.int16], waveform_pb2.I16AnalogWaveform]
):
    """A converter for AnalogWaveform types with 16-bit integer data."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return AnalogWaveform

    @property
    def python_typename(self) -> str:
        """The Python type name that this converter handles."""
        base_typename = super().python_typename
        return f"{base_typename}[int16]"

    @property
    def protobuf_message(self) -> Type[waveform_pb2.I16AnalogWaveform]:
        """The type-specific protobuf message for the Python type."""
        return waveform_pb2.I16AnalogWaveform

    def to_protobuf_message(
        self, python_value: AnalogWaveform[np.int16]
    ) -> waveform_pb2.I16AnalogWaveform:
        """Convert the Python AnalogWaveform to a protobuf Int16AnalogWaveformConverter."""
        return waveform_conversion.int16_analog_waveform_to_protobuf(python_value)

    def to_python_value(
        self, protobuf_message: waveform_pb2.I16AnalogWaveform
    ) -> AnalogWaveform[np.int16]:
        """Convert the protobuf Int16AnalogWaveformConverter to a Python AnalogWaveform."""
        return waveform_conversion.int16_analog_waveform_from_protobuf(protobuf_message)


class DoubleComplexWaveformConverter(
    Converter[ComplexWaveform[np.complex128], waveform_pb2.DoubleComplexWaveform]
):
    """A converter for complex waveform types with 64-bit real and imaginary data."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return ComplexWaveform

    @property
    def python_typename(self) -> str:
        """The Python type name that this converter handles."""
        base_typename = super().python_typename
        return f"{base_typename}[complex128]"

    @property
    def protobuf_message(self) -> Type[waveform_pb2.DoubleComplexWaveform]:
        """The type-specific protobuf message for the Python type."""
        return waveform_pb2.DoubleComplexWaveform

    def to_protobuf_message(
        self, python_value: ComplexWaveform[np.complex128]
    ) -> waveform_pb2.DoubleComplexWaveform:
        """Convert the Python ComplexWaveform to a protobuf DoubleComplexWaveform."""
        return waveform_conversion.float64_complex_waveform_to_protobuf(python_value)

    def to_python_value(
        self, protobuf_message: waveform_pb2.DoubleComplexWaveform
    ) -> ComplexWaveform[np.complex128]:
        """Convert the protobuf DoubleComplexWaveform to a Python ComplexWaveform."""
        return waveform_conversion.float64_complex_waveform_from_protobuf(protobuf_message)


class Int16ComplexWaveformConverter(
    Converter[ComplexWaveform[ComplexInt32Base], waveform_pb2.I16ComplexWaveform]
):
    """A converter for complex waveform types with 16-bit real and imaginary data."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return ComplexWaveform

    @property
    def python_typename(self) -> str:
        """The Python type name that this converter handles."""
        base_typename = super().python_typename
        # Use the string representation of ComplexInt32DType
        return f"{base_typename}[[('real', '<i2'), ('imag', '<i2')]]"

    @property
    def protobuf_message(self) -> Type[waveform_pb2.I16ComplexWaveform]:
        """The type-specific protobuf message for the Python type."""
        return waveform_pb2.I16ComplexWaveform

    def to_protobuf_message(
        self, python_value: ComplexWaveform[ComplexInt32Base]
    ) -> waveform_pb2.I16ComplexWaveform:
        """Convert the Python ComplexWaveform to a protobuf I16ComplexWaveform."""
        return waveform_conversion.int16_complex_waveform_to_protobuf(python_value)

    def to_python_value(
        self, protobuf_message: waveform_pb2.I16ComplexWaveform
    ) -> ComplexWaveform[ComplexInt32Base]:
        """Convert the protobuf I16ComplexWaveform to a Python ComplexWaveform."""
        return waveform_conversion.int16_complex_waveform_from_protobuf(protobuf_message)


class DigitalWaveformConverter(Converter[DigitalWaveform[Any], waveform_pb2.DigitalWaveform]):
    """A converter for digital waveform types."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return DigitalWaveform

    @property
    def protobuf_message(self) -> Type[waveform_pb2.DigitalWaveform]:
        """The type-specific protobuf message for the Python type."""
        return waveform_pb2.DigitalWaveform

    def to_protobuf_message(
        self, python_value: DigitalWaveform[Any]
    ) -> waveform_pb2.DigitalWaveform:
        """Convert the Python DigitalWaveform to a protobuf DigitalWaveform."""
        return waveform_conversion.digital_waveform_to_protobuf(python_value)

    def to_python_value(
        self, protobuf_message: waveform_pb2.DigitalWaveform
    ) -> DigitalWaveform[Any]:
        """Convert the protobuf DigitalWaveform to a Python DigitalWaveform."""
        return waveform_conversion.digital_waveform_from_protobuf(protobuf_message)


class DoubleSpectrumConverter(Converter[Spectrum[np.float64], waveform_pb2.DoubleSpectrum]):
    """A converter for spectrums with float64 data."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return Spectrum

    @property
    def protobuf_message(self) -> Type[waveform_pb2.DoubleSpectrum]:
        """The type-specific protobuf message for the Python type."""
        return waveform_pb2.DoubleSpectrum

    def to_protobuf_message(
        self, python_value: Spectrum[np.float64]
    ) -> waveform_pb2.DoubleSpectrum:
        """Convert the Python Spectrum to a protobuf DoubleSpectrum."""
        return waveform_conversion.float64_spectrum_to_protobuf(python_value)

    def to_python_value(
        self, protobuf_message: waveform_pb2.DoubleSpectrum
    ) -> Spectrum[np.float64]:
        """Convert the protobuf DoubleSpectrum to a Python Spectrum."""
        return waveform_conversion.float64_spectrum_from_protobuf(protobuf_message)


class BTDateTimeConverter(Converter[bt.DateTime, precision_timestamp_pb2.PrecisionTimestamp]):
    """A converter for bintime.DateTime types."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return bt.DateTime

    @property
    def protobuf_message(self) -> Type[precision_timestamp_pb2.PrecisionTimestamp]:
        """The type-specific protobuf message for the Python type."""
        return precision_timestamp_pb2.PrecisionTimestamp

    def to_protobuf_message(
        self, python_value: bt.DateTime
    ) -> precision_timestamp_pb2.PrecisionTimestamp:
        """Convert the Python DateTime to a protobuf PrecisionTimestamp."""
        return precision_timestamp_conversion.bintime_datetime_to_protobuf(python_value)

    def to_python_value(
        self, protobuf_message: precision_timestamp_pb2.PrecisionTimestamp
    ) -> bt.DateTime:
        """Convert the protobuf PrecisionTimestamp to a Python DateTime."""
        return precision_timestamp_conversion.bintime_datetime_from_protobuf(protobuf_message)


class HTDateTimeConverter(Converter[ht.datetime, precision_timestamp_pb2.PrecisionTimestamp]):
    """A converter for hightime.datetime objects.

    .. note:: The nipanel package will always convert PrecisionTimestamp messages to
        bintime.DateTime objects using BTDateTimeConverter. To use hightime.datetime
        values in a panel, you must pass a hightime.datetime value for the default_value
        parameter of the get_value() method on the panel.
    """

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return ht.datetime

    @property
    def protobuf_message(self) -> Type[precision_timestamp_pb2.PrecisionTimestamp]:
        """The type-specific protobuf message for the Python type."""
        return precision_timestamp_pb2.PrecisionTimestamp

    @property
    def protobuf_typename(self) -> str:
        """The protobuf name for the type."""
        # Override the base class here because there can only be one converter that
        # converts PrecisionTimestamp objects. Since there are two converters that convert
        # to PrecisionTimestamp, we have to choose one to handle conversion from protobuf.
        # For the purposes of nipanel, we'll convert PrecisionTimestamp messages to
        # bintime.DateTime. See BTDateTimeConverter.
        return "PrecisionTimestamp_Placeholder"

    def to_protobuf_message(
        self, python_value: ht.datetime
    ) -> precision_timestamp_pb2.PrecisionTimestamp:
        """Convert the Python DateTime to a protobuf PrecisionTimestamp."""
        return precision_timestamp_conversion.hightime_datetime_to_protobuf(python_value)

    def to_python_value(
        self, protobuf_message: precision_timestamp_pb2.PrecisionTimestamp
    ) -> ht.datetime:
        """Convert the protobuf PrecisionTimestamp to a Python DateTime."""
        return precision_timestamp_conversion.hightime_datetime_from_protobuf(protobuf_message)


class ScalarConverter(Converter[Scalar[_AnyScalarType], scalar_pb2.Scalar]):
    """A converter for Scalar objects."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return Scalar

    @property
    def protobuf_message(self) -> Type[scalar_pb2.Scalar]:
        """The type-specific protobuf message for the Python type."""
        return scalar_pb2.Scalar

    def to_protobuf_message(self, python_value: Scalar[_AnyScalarType]) -> scalar_pb2.Scalar:
        """Convert the Python Scalar to a protobuf scalar_pb2.Scalar."""
        return scalar_conversion.scalar_to_protobuf(python_value)

    def to_python_value(self, protobuf_message: scalar_pb2.Scalar) -> Scalar[_AnyScalarType]:
        """Convert the protobuf message to a Python Scalar."""
        return scalar_conversion.scalar_from_protobuf(protobuf_message)


class VectorConverter(Converter[Vector[_AnyScalarType], vector_pb2.Vector]):
    """A converter for Vector objects."""

    @property
    def python_type(self) -> type:
        """The Python type that this converter handles."""
        return Vector

    @property
    def protobuf_message(self) -> Type[vector_pb2.Vector]:
        """The type-specific protobuf message for the Python type."""
        return vector_pb2.Vector

    def to_protobuf_message(self, python_value: Vector[Any]) -> vector_pb2.Vector:
        """Convert the Python Vector to a protobuf vector_pb2.Vector."""
        return vector_conversion.vector_to_protobuf(python_value)

    def to_python_value(self, protobuf_message: vector_pb2.Vector) -> Vector[_AnyScalarType]:
        """Convert the protobuf message to a Python Vector."""
        return vector_conversion.vector_from_protobuf(protobuf_message)



"""The NI Panel."""

from importlib.metadata import version

from nipanel._panel_value_accessor import PanelValueAccessor
from nipanel._streamlit_panel import StreamlitPanel
from nipanel._streamlit_panel_initializer import (
    create_streamlit_panel,
    get_streamlit_panel_accessor,
)

__all__ = [
    "create_streamlit_panel",
    "get_streamlit_panel_accessor",
    "PanelValueAccessor",
    "StreamlitPanel",
]

# Hide that it was defined in a helper file
PanelValueAccessor.__module__ = __name__
StreamlitPanel.__module__ = __name__

__version__ = version(__name__)
"""nipanel version string."""



"""A set of checkboxes for selecting Flag enum values."""

from enum import Flag
from typing import TypeVar, Callable, Optional

import streamlit as st

from nipanel._panel_value_accessor import PanelValueAccessor

TFlagType = TypeVar("TFlagType", bound=Flag)


def flag_checkboxes(
    panel: PanelValueAccessor,
    label: str,
    value: TFlagType,
    key: str,
    disabled_values: Optional[TFlagType] = None,
    label_formatter: Callable[[Flag], str] = lambda x: str(x.name),
) -> TFlagType:
    """Create a set of checkboxes for a Flag enum.

    This will display a checkbox for each individual flag value in the enum. When checkboxes
    are selected or deselected, the combined Flag value will be stored in the panel under
    the specified key.

    Args:
        panel: The panel
        label: Label to display above the checkboxes
        value: The default Flag enum value (also determines the specific Flag enum type)
        key: Key to use for storing the Flag value in the panel
        disabled_values: A Flag enum value indicating which flags should be disabled.
                         If None or flag_type(0), no checkboxes are disabled.
        label_formatter: Function that formats the flag to a string for display. Default
                         uses flag.name.

    Returns:
        The selected Flag enum value with all selected flags combined
    """
    flag_type = type(value)
    if not issubclass(flag_type, Flag):
        raise TypeError(f"Expected a Flag enum type, got {type(value)}")

    st.markdown(f"<small>{label}:</small>", unsafe_allow_html=True)

    # Get all individual flag values (skip composite values and zero value)
    flag_values = [
        flag for flag in flag_type if flag.value & (flag.value - 1) == 0 and flag.value != 0
    ]

    # Create a container for flag checkboxes
    flag_container = st.container(border=True)

    # Use the provided value as the initial state for selected flags
    selected_flags = value

    # Create a checkbox for each flag
    for flag in flag_values:
        is_selected = bool(selected_flags & flag)

        is_disabled = False
        if disabled_values is not None:
            is_disabled = bool(disabled_values & flag)

        if flag_container.checkbox(
            label=label_formatter(flag),
            value=is_selected,
            key=f"{key}_{flag.name}",
            disabled=is_disabled,
        ):
            selected_flags |= flag
        else:
            selected_flags &= ~flag

    # Store the selected flags in the panel
    panel.set_value_if_changed(key, selected_flags)
    return selected_flags
